/****************************************************************/
/*                                                              */
/*      kpl.cpp                                                 */
/*                                                              */
/*      This file is a part of kpl - a program for graphical    */
/*      presentation of data sets and functions.                */
/*                                                              */
/*      Copyright (C) 1999 Werner Stille                        */
/*      <stille@uni-freiburg.de>                                */
/*                                                              */
/*      Released under the GPL; see file LICENSE for details.   */
/*                                                              */
/****************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <qprinter.h>
#include <qkeycode.h>
#include <qpaintdevicemetrics.h>
#include <qfileinfo.h>
#include <qspinbox.h>
#include <qscrollbar.h>
#include <kapp.h>
#include <kfm.h>
#include <kfiledialog.h> 
#include <kiconloader.h> 
#include <kmsgbox.h>
#include <dlfcn.h>
#include <stdio.h>
#include "kpl.h"
#include "kplwidget.h"
#include "refreshdlg.h"
#include "zoomdlg.h"
#include "itemdlg.h"
#include "pslib.h"

#if (KDE_VERSION_MAJOR > 0) && (KDE_VERSION_MINOR > 0)
  #define USE_KACCEL
  #include <kaccel.h>
#else
  #include <kstdaccel.h>
#endif

#ifndef min
  #define min(x,y) ((x) < (y) ? (x) : (y))
#endif
#ifndef max
  #define max(x,y) ((x) > (y) ? (x) : (y))
#endif

const char KPl::itemType[6][11] = {"", "FRAMEITEM", "ARRAYITEM", "FUNITEM",
                                   "PARFUNITEM", "LEGENDITEM"};
const int KPl::iSize[6] = {0, sizeof(FRAMEDATA), sizeof(ARRAYDATA),
                           sizeof(FUNDATA), sizeof(PARFUNDATA),
                           sizeof(LEGENDDATA)};
QList<KPl> KPl::windowList;

/****************************************************************/
/*                                                              */
/*      freeX(x)                                                */
/*        deallocates array storage                             */
/*                                                              */
/*      x:        pointer to array                              */
/*                                                              */
/****************************************************************/

void KPl::freeX(double ***x)
{
  if (*x) {
    free((*x)[0]);
    free(*x);
    *x = 0;
  }
}

/****************************************************************/
/*                                                              */
/*      freeXCond(ad)                                           */
/*        deallocates array storage of a structure, when this   */
/*        is the only one using the array                       */
/*                                                              */
/*      ad:       pointer to array data structure               */
/*                                                              */
/****************************************************************/

void KPl::freeXCond(PARRAYDATA ad)
{
  int i, del = TRUE;

  for (i = 0; i < nItems; i++)
    if ((itd[i].stru != ad) && (itd[i].ityp == ARRAYITEM))
      if (((PARRAYDATA) itd[i].stru)->x == ad->x)
        del = 0;
  if (del)
    freeX(&ad->x);
}

/****************************************************************/
/*                                                              */
/*      freeItd(iti)                                            */
/*        deallocates storage of an plot item                   */
/*                                                              */
/*      iti:       pointer zo structure                         */
/*                                                              */
/****************************************************************/

void KPl::freeItd(PITEMDATA iti)
{
  switch (iti->ityp) {
    case ARRAYITEM:
      if (((PARRAYDATA) iti->stru)->x)
        freeXCond((PARRAYDATA) iti->stru);
      break;
    case FUNITEM:
      if (((PFUNDATA) iti->stru)->hmod)
        dlclose(((PFUNDATA) iti->stru)->hmod);
      break;
    case PARFUNITEM:
      if (((PPARFUNDATA) iti->stru)->hmodx)
        dlclose(((PPARFUNDATA) iti->stru)->hmodx);
      if (((PPARFUNDATA) iti->stru)->hmody)
        dlclose(((PPARFUNDATA) iti->stru)->hmody);
  }
  free(iti->stru);
}

/****************************************************************/
/*                                                              */
/*      freeAllItd()                                            */
/*        deallocates storage of all structures                 */
/*                                                              */
/****************************************************************/

void KPl::freeAllItd()
{
  int i;

  for (i = nItems - 1; i >= 0; i--) {
    freeItd(&itd[i]);
    nItems--;
  }
  if (itd) {
    free(itd);
    itd = 0;
  }
  changed = 0;
}

/****************************************************************/
/*                                                              */
/*      readFile(filename, x, ncols, nrows)                     */
/*        reads double values from a file into an array         */
/*                                                              */
/*      filename: string containing file path                   */
/*      x:        pointer to two-dimensional double array       */
/*      ncols:    pointer to number of colums (return value)    */
/*      nrows:    pointer to number of lines (return value)     */
/*                                                              */
/*      returns:  number of lines                               */
/*                0 for error                                   */
/*                                                              */
/****************************************************************/

int KPl::readFile(QString filename, double ***x, int *ncols, int *nrows)
{
  const char delim[] = " ,\t";
  int i, j, k, np;
  double *x1, *x2 = 0;
  char s[256], *s1;
  FILE *f;

  freeX(x);
  *ncols = *nrows = 0;
  if ((f = fopen(filename, "r"))) {
    do
      if ((i = (fgets(s, 256, f) != 0)))
        i = (s[0] == '#') ? 1 : 2;
    while (i == 1);
    if (i) {
      s1 = strtok(s, delim);
      while (s1) {
        (*ncols)++;
        s1 = strtok(0, delim);
      }
      if (*ncols) {
        rewind(f);
        np = 0;
        k = 0;
        while (fgets(s, 256, f))
          if (s[0] != '#') {
            (*nrows)++;
            np += *ncols;
            x2 = (double *) realloc(x2, np * sizeof(double));
            x2[k++] = ((s1 = strtok(s, delim))) ? atof(s1) : 0.0;
            for (i = 1; i < *ncols; i++)
              x2[k++] = ((s1 = strtok(0, delim))) ? atof(s1) : 0.0;
          }
        x1 = (double *) malloc(np * sizeof(double));
        *x = (double **) malloc(*ncols * sizeof(double *));
        for (i = 0; i < *ncols; i++)
          (*x)[i] = &x1[i * *nrows];
        k = 0;
        for (j = 0; j < *nrows; j++)
          for (i = 0; i < *ncols; i++)
            (*x)[i][j] = x2[k++];
        free(x2);
      }
    }
    fclose(f);
  }
  return *nrows;
}

/****************************************************************/
/*                                                              */
/*      iniStru(itd, ityp)                                      */
/*        initializes structure                                 */
/*                                                              */
/*      itd:      pointer to structure                          */
/*      ityp:     type of structure                             */
/*                                                              */
/****************************************************************/

void KPl::iniStru(PITEMDATA itd, int ityp)
{
  double dx = aut.xlAuto / 15.0;

  itd->ityp = ityp;
  itd->stru = calloc(1, iSize[ityp]);
  switch (ityp) {
    case FRAMEITEM:
      ((PFRAMEDATA) itd->stru)->xf = bas.xf;
      ((PFRAMEDATA) itd->stru)->yf = bas.yf;
      ((PFRAMEDATA) itd->stru)->colFrame = bas.colFrame;
      ((PFRAMEDATA) itd->stru)->colGrid = bas.colGrid;
      ((PFRAMEDATA) itd->stru)->xw0 = aut.x0Auto;
      ((PFRAMEDATA) itd->stru)->xw = aut.xlAuto;
      ((PFRAMEDATA) itd->stru)->yw0 = aut.y0Auto;
      ((PFRAMEDATA) itd->stru)->yw = aut.ylAuto;
      ((PFRAMEDATA) itd->stru)->xma = aut.xlAuto;
      ((PFRAMEDATA) itd->stru)->yma = aut.ylAuto;
      ((PFRAMEDATA) itd->stru)->xtic = 5.0;
      ((PFRAMEDATA) itd->stru)->ytic = 2.0;
      ((PFRAMEDATA) itd->stru)->relsiz = 1.0;
      ((PFRAMEDATA) itd->stru)->mticx = 5;
      ((PFRAMEDATA) itd->stru)->mticy = 2;
      ((PFRAMEDATA) itd->stru)->ndigx = -1;
      ((PFRAMEDATA) itd->stru)->ndigy = -1;
      ((PFRAMEDATA) itd->stru)->gridmode = aut.autoGrid;
      strcpy(((PFRAMEDATA) itd->stru)->sx, aut.sxAuto);
      strcpy(((PFRAMEDATA) itd->stru)->sy, aut.syAuto);
      strcpy(((PFRAMEDATA) itd->stru)->sh, aut.shAuto);
      break;
    case ARRAYITEM:
      ((PARRAYDATA) itd->stru)->color = bas.colData;
      ((PARRAYDATA) itd->stru)->fx = 1.0;
      ((PARRAYDATA) itd->stru)->fy = 1.0;
      ((PARRAYDATA) itd->stru)->ix = aut.ixAuto;
      ((PARRAYDATA) itd->stru)->iy = aut.iyAuto;
      ((PARRAYDATA) itd->stru)->ie = aut.ieAuto;
      ((PARRAYDATA) itd->stru)->symb = aut.autoSymb;
      ((PARRAYDATA) itd->stru)->errbars = aut.autoErr;
      break;
    case FUNITEM:
      ((PFUNDATA) itd->stru)->color = bas.colData;
      ((PFUNDATA) itd->stru)->fx = 1.0;
      ((PFUNDATA) itd->stru)->fy = 1.0;
      ((PFUNDATA) itd->stru)->symb = aut.autoSymb;
      break;
    case PARFUNITEM:
      ((PPARFUNDATA) itd->stru)->color = bas.colData;
      ((PPARFUNDATA) itd->stru)->fx = 1.0;
      ((PPARFUNDATA) itd->stru)->fy = 1.0;
      ((PPARFUNDATA) itd->stru)->symb = aut.autoSymb;
      break;
    case LEGENDITEM:
      ((PLEGENDDATA) itd->stru)->colsymb = bas.colData;
      ((PLEGENDDATA) itd->stru)->coltext = bas.colFrame;
      ((PLEGENDDATA) itd->stru)->x = aut.x0Auto + dx;
      ((PLEGENDDATA) itd->stru)->y = aut.y0Auto + aut.ylAuto
        - (++nyLeg) * dx;
      ((PLEGENDDATA) itd->stru)->xtext = dx;
      ((PLEGENDDATA) itd->stru)->ytext = -0.18 * dx;
      ((PLEGENDDATA) itd->stru)->relsiz = 1.0;
      ((PLEGENDDATA) itd->stru)->symb = aut.autoSymb;
  }
}

/****************************************************************/
/*                                                              */
/*      autoStruc(fd, ad)                                       */
/*        sets array and data structures by autoscaling         */
/*                                                              */
/*      returns:  !0 for error                                  */
/*                                                              */
/****************************************************************/

int KPl::autoStruc(PFRAMEDATA fd, PARRAYDATA ad)
{
  if ((ad->x) && (aut.ixAuto < ad->ncols) && (aut.iyAuto < ad->ncols)) {
    if (autoFl(&fd->xmi, &fd->xma, &fd->xtic, &fd->mticx, &fd->ndigx, &fd->iex,
               &ad->fx, ad->x[aut.ixAuto], ad->nrows, aut.autoNorm, 0.2))
      return TRUE;
    if (autoFl(&fd->ymi, &fd->yma, &fd->ytic, &fd->mticy, &fd->ndigy, &fd->iey,
               &ad->fy, ad->x[aut.iyAuto], ad->nrows, aut.autoNorm,
               0.2 * aut.xlAuto / aut.ylAuto))
      return TRUE;
    if (aut.autohPath)
      strncpy(fd->sh, ad->path, sizeof(aut.shAuto) - 1);
    ad->ix = aut.ixAuto;
    ad->iy = aut.iyAuto;
    ad->ie = aut.ieAuto;
    ad->symb = aut.autoSymb;
    ad->istart = 0;
    ad->n = ad->nrows;
    ad->errbars = aut.autoErr & (aut.ieAuto < ad->ncols);
    return 0;
  }
  return TRUE;
}

/****************************************************************/
/*                                                              */
/*      updWin()                                                */
/*        updates menus and windows                             */
/*                                                              */
/****************************************************************/

void KPl::updWin()
{
  QString s;

  if (!currentPath.isEmpty())
    setCaption(s.sprintf("%s - %s",
                         (const char *) currentPath,
                         (const char *) kapp->appName()));
  else
    setCaption(kapp->appName());
  filePopup->setItemEnabled(IDM_PLOT, nItems);
  filePopup->setItemEnabled(IDM_PRINT, nItems);
  toolBar()->setItemEnabled(IDM_PRINT, nItems);
  filePopup->setItemEnabled(IDM_PS, nItems);
  filePopup->setItemEnabled(IDM_SAVE, changed);
  toolBar()->setItemEnabled(IDM_SAVE, changed);
  filePopup->setItemEnabled(IDM_SAVEAS, nItems);
  plot = optionsPopup->isItemChecked(IDM_AUTOPLOT) && nItems;
  if (plot && isVisible())
    paintWidget();
}

/****************************************************************/
/*                                                              */
/*      getFun(path, name, hmod, fkt)                           */
/*        loads shared library and determines function          */
/*        addresses                                             */
/*                                                              */
/*      path:     string containing file path                   */
/*      name:     string containing function name               */
/*      hmod:     pointer to module handle                      */
/*      fkt:      pointer to function                           */
/*                                                              */
/****************************************************************/

void KPl::getFun(QString path, QString name, void **hmod,
                   double (**fkt)(double, double *))
{
  char *error;
  QString msg;

  *hmod = 0;
  *fkt = 0;
  if (!path.isEmpty()) {
    if (!(*hmod = dlopen(path, RTLD_NOW)))
      KMsgBox::message(this, i18n("Error"),
                       msg.sprintf(i18n("while trying to open\nfile %s:\n%s"),
                                   (const char *) path, dlerror()), 2);
    else
      if (!name.isEmpty())
        *fkt = (double (*)(double, double *)) dlsym(*hmod, name);
        if ((error = (char *) dlerror()) != 0)
          KMsgBox::message(this, i18n("Error"),
                           msg.sprintf(i18n("while trying to determine the "
                                            "address of\nfunction %s:\n%s"),
                                       (const char *) name, error), 2);
  }
}

/****************************************************************/
/*                                                              */
/*      readPlo(f)                                              */
/*        reads PLO file and initializes structures             */
/*                                                              */
/*      f:        pointer to stream opened for write access     */
/*                                                              */
/*      returns:  number of plot items                          */
/*                                                              */
/****************************************************************/

int KPl::readPlo(FILE *f)
{
  const char framename[][9] = {"ndigx", "ndigy", "mticx", "mticy", "gridmode",
                               "iex", "iey", "colframe", "colgrid",
                               "xf", "yf", "xw0", "xw", "yw0", "yw",
                               "xmi", "xma", "ymi", "yma", "xtic", "ytic",
                               "relsiz", "sx", "sy", "sh"};
  const char arrayname[][8] = {"ix", "iy", "ie", "istart", "n", "symb",
                               "errbars", "color", "fx", "fy", "path"};
  const char funname[][6] = {"symb", "color", "xfmin", "xfmax", "dx", "fx",
                             "fy", "p", "name", "path"};
  const char parfunname[][6] = {"symb", "color", "tmin", "tmax", "dt", "fx",
                                "fy", "px", "py", "namex", "namey",
                                "pathx", "pathy"};
  const char legendname[][8] = {"symb", "colsymb", "coltext", "x", "y",
                                "xtext", "ytext", "relsiz", "s"};
  int i = -99, j, k;
  char line[256], *p;

  while (fgets(line, sizeof(line), f)) {
    if ((p = strchr(line, '\r')) != 0)
      *p = '\0';
    if ((p = strchr(line, '\n')) != 0)
      *p = '\0';
    if (line[0] == '\0')
      continue;
    if (line[0] == '[') {
      if ((p = strchr(line + 1, ']')) != 0) {
        *p = '\0';
        if (!strcmp(line + 1, "Items")) {
          if (i == -99)
            i = -1;
          else {
            nItems = 0;
            return nItems;
          }
        } else {
          if (!strncmp(line + 1, "Item ", 5))
            if (i == -99)
              return nItems;
            else {
              sscanf(line + 6, "%i", &i);
              if ((i < 0) || (i >= nItems)) {
                nItems = 0;
                return nItems;
              }
            }
        }
      }
    } else {
      if (i == -99)
        return nItems;
      strtok(line, "=");
      p = line + strlen(line) + 1;
      if (i == -1)
        if (strcmp(line, "nitems")) {
          nItems = 0;
          return nItems;
        } else {
          sscanf(p, "%i", &nItems);
          if (nItems <= 0) {
            nItems = 0;
            return nItems;
          } else
            itd = (PITEMDATA) calloc(nItems, sizeof(ITEMDATA));
        }
      else {
        j = strcmp(line, "Type");
        if (itd[i].ityp) {
          if (j) {
            switch (itd[i].ityp) {
              case FRAMEITEM:
                for (j = 0; j < 25; j++)
                  if (!strcmp(line, framename[j]))
                    break;
                if (j >= 25) {
                  nItems = 0;
                  return nItems;
                }
                if (j < 7) {
                  sscanf(p, "%i", &((PFRAMEDATA) itd[i].stru)->ndigx + j);
                  break;
                }
                if (j < 9) {
                  sscanf(p, "%lx", &((PFRAMEDATA)
                         itd[i].stru)->colFrame + j - 7);
                  break;
                }
                if (j < 22) {
                  sscanf(p, "%lg", &((PFRAMEDATA) itd[i].stru)->xf + j - 9);
                  break;
                }
                strcpy(((PFRAMEDATA) itd[i].stru)->sx + 80 * (j - 22), p);
                break;
              case ARRAYITEM:
                for (j = 0; j < 11; j++)
                  if (!strcmp(line, arrayname[j]))
                    break;
                if (j >= 25) {
                  nItems = 0;
                  return nItems;
                }
                if (j < 7) {
                  sscanf(p, "%i", &((PARRAYDATA) itd[i].stru)->ix + j);
                  break;
                }
                if (j < 8) {
                  sscanf(p, "%lx", &((PARRAYDATA) itd[i].stru)->color);
                  break;
                }
                if (j < 10) {
                  sscanf(p, "%lg", &((PARRAYDATA) itd[i].stru)->fx + j - 8);
                  break;
                }
                strcpy(((PARRAYDATA) itd[i].stru)->path, p);
                break;
              case FUNITEM:
                for (j = 0; j < 10; j++)
                  if (!strcmp(line, funname[j]))
                    break;
                if (j >= 10) {
                  nItems = 0;
                  return nItems;
                }
                if (j == 0) {
                  sscanf(p, "%i", &((PFUNDATA) itd[i].stru)->symb);
                  break;
                }
                if (j == 1) {
                  sscanf(p, "%lx", &((PFUNDATA) itd[i].stru)->color);
                  break;
                }
                if (j < 7) {
                  sscanf(p, "%lg", &((PFUNDATA) itd[i].stru)->xfmin + j - 2);
                  break;
                }
                if (j == 7) {
                  strtok(p, " ");
                  for (k = 0; k < 20; k++) {
                    sscanf(p, "%lg", &((PFUNDATA) itd[i].stru)->p[k]);
                    p = strtok(0, " ");
                  }
                  break;
                }
                strcpy(((PFUNDATA) itd[i].stru)->name + 100 * (j - 8), p);
                break;
              case PARFUNITEM:
                for (j = 0; j < 13; j++)
                  if (!strcmp(line, parfunname[j]))
                    break;
                if (j >= 13) {
                  nItems = 0;
                  return nItems;
                }
                if (j == 0) {
                  sscanf(p, "%i", &((PPARFUNDATA) itd[i].stru)->symb);
                  break;
                }
                if (j == 1) {
                  sscanf(p, "%lx", &((PPARFUNDATA) itd[i].stru)->color);
                  break;
                }
                if (j < 7) {
                  sscanf(p, "%lg", &((PPARFUNDATA) itd[i].stru)->tmin + j - 2);
                  break;
                }
                if (j == 7) {
                  strtok(p, " ");
                  for (k = 0; k < 20; k++) {
                    sscanf(p, "%lg", &((PPARFUNDATA) itd[i].stru)->px[k]);
                    p = strtok(0, " ");
                  }
                  break;
                }
                if (j == 8) {
                  strtok(p, " ");
                  for (k = 0; k < 20; k++) {
                    sscanf(p, "%lg", &((PPARFUNDATA) itd[i].stru)->py[k]);
                    p = strtok(0, " ");
                  }
                  break;
                }
                if (j < 11) {
                  strcpy(((PPARFUNDATA) itd[i].stru)->namex + 100 * (j - 9),
                         p);
                  break;
                }
                strcpy(((PPARFUNDATA) itd[i].stru)->pathx + 260 * (j - 11),
                       p);
                break;
              case LEGENDITEM:
                for (j = 0; j < 9; j++)
                  if (!strcmp(line, legendname[j]))
                    break;
                if (j >= 9) {
                  nItems = 0;
                  return nItems;
                }
                if (j == 0) {
                  sscanf(p, "%i", &((PLEGENDDATA) itd[i].stru)->symb);
                  break;
                }
                if (j < 3) {
                  sscanf(p, "%lx", &((PLEGENDDATA)
                         itd[i].stru)->colsymb + j - 1);
                  break;
                }
                if (j < 8) {
                  sscanf(p, "%lg", &((PLEGENDDATA) itd[i].stru)->x + j - 3);
                  break;
                }
                strcpy(((PLEGENDDATA) itd[i].stru)->s, p);
                break;
            }
          } else {
            nItems = 0;
            return nItems;
          }
        } else {
          if (j) {
            nItems = 0;
            return nItems;
          } else {
            for (itd[i].ityp = 1; itd[i].ityp < 6; itd[i].ityp++)
              if (!strcmp(p, itemType[itd[i].ityp]))
                break;
            if (itd[i].ityp >= 6) {
              nItems = 0;
              return nItems;
            }
            itd[i].stru = calloc(iSize[itd[i].ityp], 1);
          }
        }
      }
    }
  }
  fclose(f);
  return nItems;
}

/****************************************************************/
/*                                                              */
/*      writePlo(f)                                             */
/*        writes PLO file                                       */
/*                                                              */
/*      f:        pointer to stream opened for write access     */
/*                                                              */
/****************************************************************/

void KPl::writePlo(FILE *f)
{
  int i, j;

  fprintf(f, "[Items]\nnitems=%i\n", nItems);
  for (i = 0; i < nItems; i++) {
    fprintf(f, "\n[Item %i]\nType=%s\n", i, itemType[itd[i].ityp]);
    switch (itd[i].ityp) {
      case FRAMEITEM:
        fprintf(f, "ndigx=%i\nndigy=%i\nmticx=%i\nmticy=%i\ngridmode=%i\n"
                "iex=%i\niey=%i\ncolframe=%lx\ncolgrid=%lx\n"
                "xf=%g\nyf=%g\nxw0=%g\nxw=%g\nyw0=%g\nyw=%g\n"
                "xmi=%g\nxma=%g\nymi=%g\nyma=%g\nxtic=%g\nytic=%g\n"
                "relsiz=%g\nsx=%s\nsy=%s\nsh=%s\n",
                ((PFRAMEDATA) itd[i].stru)->ndigx,
                ((PFRAMEDATA) itd[i].stru)->ndigy,
                ((PFRAMEDATA) itd[i].stru)->mticx,
                ((PFRAMEDATA) itd[i].stru)->mticy,
                ((PFRAMEDATA) itd[i].stru)->gridmode,
                ((PFRAMEDATA) itd[i].stru)->iex,
                ((PFRAMEDATA) itd[i].stru)->iey,
                ((PFRAMEDATA) itd[i].stru)->colFrame,
                ((PFRAMEDATA) itd[i].stru)->colGrid,
                ((PFRAMEDATA) itd[i].stru)->xf,
                ((PFRAMEDATA) itd[i].stru)->yf,
                ((PFRAMEDATA) itd[i].stru)->xw0,
                ((PFRAMEDATA) itd[i].stru)->xw,
                ((PFRAMEDATA) itd[i].stru)->yw0,
                ((PFRAMEDATA) itd[i].stru)->yw,
                ((PFRAMEDATA) itd[i].stru)->xmi,
                ((PFRAMEDATA) itd[i].stru)->xma,
                ((PFRAMEDATA) itd[i].stru)->ymi,
                ((PFRAMEDATA) itd[i].stru)->yma,
                ((PFRAMEDATA) itd[i].stru)->xtic,
                ((PFRAMEDATA) itd[i].stru)->ytic,
                ((PFRAMEDATA) itd[i].stru)->relsiz,
                ((PFRAMEDATA) itd[i].stru)->sx,
                ((PFRAMEDATA) itd[i].stru)->sy,
                ((PFRAMEDATA) itd[i].stru)->sh);
        break;
      case ARRAYITEM:
        fprintf(f, "ix=%i\niy=%i\nie=%i\nistart=%i\nn=%i\nsymb=%i\n"
                "errbars=%i\ncolor=%lx\nfx=%g\nfy=%g\npath=%s\n",
                ((PARRAYDATA) itd[i].stru)->ix,
                ((PARRAYDATA) itd[i].stru)->iy,
                ((PARRAYDATA) itd[i].stru)->ie,
                ((PARRAYDATA) itd[i].stru)->istart,
                ((PARRAYDATA) itd[i].stru)->n,
                ((PARRAYDATA) itd[i].stru)->symb,
                ((PARRAYDATA) itd[i].stru)->errbars,
                ((PARRAYDATA) itd[i].stru)->color,
                ((PARRAYDATA) itd[i].stru)->fx,
                ((PARRAYDATA) itd[i].stru)->fy,
                ((PARRAYDATA) itd[i].stru)->path);
         break;
      case FUNITEM:
        fprintf(f, "symb=%i\ncolor=%lx\nxfmin=%g\nxfmax=%g\ndx=%g\n"
                "fx=%g\nfy=%g\np=",
                ((PFUNDATA) itd[i].stru)->symb,
                ((PFUNDATA) itd[i].stru)->color,
                ((PFUNDATA) itd[i].stru)->xfmin,
                ((PFUNDATA) itd[i].stru)->xfmax,
                ((PFUNDATA) itd[i].stru)->dx,
                ((PFUNDATA) itd[i].stru)->fx,
                ((PFUNDATA) itd[i].stru)->fy);
        for (j = 0; j < 19; j++)
          fprintf(f, "%g ",((PFUNDATA) itd[i].stru)->p[j]);
        fprintf(f, "%g\nname=%s\npath=%s\n",
                ((PFUNDATA) itd[i].stru)->p[19],
                ((PFUNDATA) itd[i].stru)->name,
                ((PFUNDATA) itd[i].stru)->path);
        break;
      case PARFUNITEM:
        fprintf(f, "symb=%i\ncolor=%lx\ntmin=%g\ntmax=%g\ndt=%g\n"
                "fx=%g\nfy=%g\npx=",
                ((PPARFUNDATA) itd[i].stru)->symb,
                ((PPARFUNDATA) itd[i].stru)->color,
                ((PPARFUNDATA) itd[i].stru)->tmin,
                ((PPARFUNDATA) itd[i].stru)->tmax,
                ((PPARFUNDATA) itd[i].stru)->dt,
                ((PPARFUNDATA) itd[i].stru)->fx,
                ((PPARFUNDATA) itd[i].stru)->fy);
        for (j = 0; j < 19; j++)
          fprintf(f, "%g ",((PPARFUNDATA) itd[i].stru)->px[j]);
        fprintf(f, "%g\npy=",((PPARFUNDATA) itd[i].stru)->px[19]);
        for (j = 0; j < 19; j++)
          fprintf(f, "%g ",((PPARFUNDATA) itd[i].stru)->py[j]);
        fprintf(f, "%g\nnamex=%s\nnamey=%s\npathx=%s\npathy=%s\n",
                ((PPARFUNDATA) itd[i].stru)->py[19],
                ((PPARFUNDATA) itd[i].stru)->namex,
                ((PPARFUNDATA) itd[i].stru)->namey,
                ((PPARFUNDATA) itd[i].stru)->pathx,
                ((PPARFUNDATA) itd[i].stru)->pathy);
        break;
      case LEGENDITEM:
        fprintf(f, "symb=%i\ncolsymb=%lx\ncoltext=%lx\nx=%g\ny=%g\n"
                "xtext=%g\nytext=%g\nrelsiz=%g\ns=%s\n",
                ((PLEGENDDATA) itd[i].stru)->symb,
                ((PLEGENDDATA) itd[i].stru)->colsymb,
                ((PLEGENDDATA) itd[i].stru)->coltext,
                ((PLEGENDDATA) itd[i].stru)->x,
                ((PLEGENDDATA) itd[i].stru)->y,
                ((PLEGENDDATA) itd[i].stru)->xtext,
                ((PLEGENDDATA) itd[i].stru)->ytext,
                ((PLEGENDDATA) itd[i].stru)->relsiz,
                ((PLEGENDDATA) itd[i].stru)->s);
    }
  }
  fclose(f);
}

/****************************************************************/
/*                                                              */
/*      getFile(path)                                           */
/*        reads file and initializes structures                 */
/*                                                              */
/*      path:     string containing file path                   */
/*                                                              */
/****************************************************************/

void KPl::getFile(QString path)
{
  int i;
  FILE *f;

  if (!path.isEmpty()) {
    QFileInfo fi(path);
    if (!fType) {
      QString ext = fi.extension();
      fType = stricmp(ext, "plo") ? DATAFILE : PLOTFILE;
    }
    if (fType == DATAFILE) {
      freeAllItd();
      nItems = 2;
      itd = (PITEMDATA) calloc(nItems, sizeof(ITEMDATA));
      iniStru(&itd[0], FRAMEITEM);
      iniStru(&itd[1], ARRAYITEM);
      strcpy(((PARRAYDATA) itd[1].stru)->path, path);
      if (!readFile(((PARRAYDATA) itd[1].stru)->path,
                    &((PARRAYDATA) itd[1].stru)->x,
                    &((PARRAYDATA) itd[1].stru)->ncols,
                    &((PARRAYDATA) itd[1].stru)->nrows)) {
        freeAllItd();
        KMsgBox::message(this, i18n("Error"),
                         i18n("while loading the file"), 2);
        currentPath = "";
        updWin();
        return;
      } else {
        changed = TRUE;
        autoStruc((PFRAMEDATA) itd[0].stru,  (PARRAYDATA) itd[1].stru);
      }
    } else {
      if ((f = fopen(path, "r"))) {
        freeAllItd();
        nItems = readPlo(f);
        if (nItems) {
          for (i = 0; i < nItems; i++) {
            switch (itd[i].ityp) {
              case ARRAYITEM:
                ((PARRAYDATA) itd[i].stru)->x = 0;
                if (!readFile(((PARRAYDATA) itd[i].stru)->path,
                              &((PARRAYDATA) itd[i].stru)->x,
                              &((PARRAYDATA) itd[i].stru)->ncols,
                              &((PARRAYDATA) itd[i].stru)->nrows))
                  KMsgBox::message(this, i18n("Error"),
                                   (((PARRAYDATA) itd[i].stru)->path), 2);
                break;
              case FUNITEM:
                getFun(((PFUNDATA) itd[i].stru)->path,
                       ((PFUNDATA) itd[i].stru)->name,
                       &((PFUNDATA) itd[i].stru)->hmod,
                       &((PFUNDATA) itd[i].stru)->fkt);
                break;
              case PARFUNITEM:
                getFun(((PPARFUNDATA) itd[i].stru)->pathx,
                       ((PPARFUNDATA) itd[i].stru)->namex,
                       &((PPARFUNDATA) itd[i].stru)->hmodx,
                       &((PPARFUNDATA) itd[i].stru)->fktx);
                getFun(((PPARFUNDATA) itd[i].stru)->pathy,
                       ((PPARFUNDATA) itd[i].stru)->namey,
                       &((PPARFUNDATA) itd[i].stru)->hmody,
                       &((PPARFUNDATA) itd[i].stru)->fkty);
            }
          }
        }
      } else {
        currentPath = "";
        updWin();
        return;
      }
    }
    currentPath = path;
    currentDir = fi.dirPath(TRUE);
  }
  updWin();
}

/****************************************************************/
/*                                                              */
/*      drawFrame(fd)                                           */
/*        draws frame, axes, and labels                         */
/*                                                              */
/*      fd:       pointer to frame structure                    */
/*                                                              */
/****************************************************************/

void KPl::drawFrame(PFRAMEDATA fd)
{
  if (fd->xw && fd->yw) {
    format(fd->xf, fd->yf);
    relsize(fd->relsiz);
    Window(fd->xw0, fd->xw, fd->yw0, fd->yw);
    scale(fd->xmi, fd->xma, fd->ymi, fd->yma);
    setdig(fd->ndigx, fd->ndigy);
    setColFrame(QColor(fd->colFrame));
    setColGrid(QColor(fd->colGrid));
    raster(fd->xtic, fd->ytic, fd->mticx, fd->mticy, fd->gridmode);
    letx(fd->sx, fd->iex);
    lety(fd->sy, fd->iey);
    leth(fd->sh);
  }
}

/****************************************************************/
/*                                                              */
/*      drawArray(ad)                                           */
/*        plots data array and error bars                       */
/*                                                              */
/*      ad:       pointer to structure containg data array      */
/*                                                              */
/****************************************************************/

void KPl::drawArray(PARRAYDATA ad)
{
  if (!ad->x)
    return;
  setColData(QColor(ad->color));
  symbol(ad->symb);
  plarry(&ad->x[ad->ix][ad->istart], &ad->x[ad->iy][ad->istart],
         ad->fx, ad->fy, ad->n);
  if (ad->errbars)
    plerry(&ad->x[ad->ix][ad->istart], &ad->x[ad->iy][ad->istart],
           &ad->x[ad->ie][ad->istart], ad->fx, ad->fy, ad->n);
  fxn = ad->fx;
  fyn = ad->fy;
}

/****************************************************************/
/*                                                              */
/*      drawFun(fd)                                             */
/*        plots functions                                       */
/*                                                              */
/*      fd:       pointer to function structure                 */
/*                                                              */
/****************************************************************/

void KPl::drawFun(PFUNDATA fd)
{
  if (fd->fkt) {
    setColData(QColor(fd->color));
    symbol(fd->symb);
    plfun(fd->xfmin, fd->xfmax, fd->dx, fd->fx, fd->fy, fd->fkt, fd->p);
  }
}

/****************************************************************/
/*                                                              */
/*      drawParFun(pfd)                                         */
/*        plots parametric functions                            */
/*                                                              */
/*      pfd:      pointer to parametric function structure      */
/*                                                              */
/****************************************************************/

void KPl::drawParFun(PPARFUNDATA pfd)
{
  if (pfd->fktx && pfd->fkty) {
    setColData(QColor(pfd->color));
    symbol(pfd->symb);
    plpfun(pfd->tmin, pfd->tmax, pfd->dt, pfd->fx, pfd->fy,
           pfd->fktx, pfd->px, pfd->fkty, pfd->py);
  }
}

/****************************************************************/
/*                                                              */
/*      drawLegend(ld)                                          */
/*        plots legend                                          */
/*                                                              */
/*      ld:       pointer to legend structure                   */
/*                                                              */
/****************************************************************/

void KPl::drawLegend(PLEGENDDATA ld)
{
  int ix, iy, n;
  double x[2], y[2], sav;

  setColData(QColor(ld->colsymb));
  symbol(ld->symb);
  if (ld->symb > 0) {
    cm2r(ld->x - 0.4 * ld->xtext, ld->y, 1.0, 1.0, &x[0], &y[0]);
    cm2r(ld->x + 0.4 * ld->xtext, ld->y, 1.0, 1.0, &x[1], &y[1]);
    n = 2;
  } else {
    cm2r(ld->x, ld->y, 1.0, 1.0, &x[0], &y[0]);
    n = 1;
  }
  plarry(x, y, 1.0, 1.0, n);
  setColFrame(QColor(ld->coltext));
  cm2r(ld->x + ld->xtext, ld->y + ld->ytext + 0.18, 1.0, 1.0, &x[0], &y[0]);
  r2i(x[0], y[0], &ix, &iy);
  sav = relsize(ld->relsiz);
  text(ld->s, ix, iy, 1);
  relsize(sav);
}

/****************************************************************/
/*                                                              */
/*      doDraw()                                                */
/*        plots frame, grid, labels, data arrays,               */
/*        ordinary and parametric functions                     */
/*                                                              */
/****************************************************************/

void KPl::doDraw()
{
  int i;

  for (i = 0; i < nItems; i++)
    switch (itd[i].ityp) {
      case FRAMEITEM:
        drawFrame((PFRAMEDATA) itd[i].stru);
        break;
      case ARRAYITEM:
        drawArray((PARRAYDATA) itd[i].stru);
        break;
      case FUNITEM:
        drawFun((PFUNDATA) itd[i].stru);
        break;
      case PARFUNITEM:
        drawParFun((PPARFUNDATA) itd[i].stru);
        break;
      case LEGENDITEM:
        drawLegend((PLEGENDDATA) itd[i].stru);
    }
}

/****************************************************************/
/*                                                              */
/*      psFrame(fd)                                             */
/*        writes frame, axes, and labels to PS file             */
/*                                                              */
/*      fd:       pointer to frame structure                    */
/*                                                              */
/****************************************************************/

void KPl::psFrame(PFRAMEDATA fd)
{
  if (fd->xw && fd->yw) {
    psFormat(fd->xf, fd->yf);
    psRelSize(fd->relsiz);
    psWindow(fd->xw0, fd->xw, fd->yw0, fd->yw);
    psScale(fd->xmi, fd->xma, fd->ymi, fd->yma);
    psSetDig(fd->ndigx, fd->ndigy);
    psSetColFrame(fd->colFrame);
    psSetColGrid(fd->colGrid);
    psRaster(fd->xtic, fd->ytic, fd->mticx, fd->mticy, fd->gridmode);
    psLetX(fd->sx, fd->iex);
    psLetY(fd->sy, fd->iey);
    psLetH(fd->sh);
  }
}

/****************************************************************/
/*                                                              */
/*      psArray(ad)                                             */
/*        writes data array and errorbars to PS file            */
/*                                                              */
/*      ad:       pointer to array structure                    */
/*                                                              */
/****************************************************************/

void KPl::psArray(PARRAYDATA ad)
{
  if (!ad->x)
    return;
  psSetColData(ad->color);
  psSymbol(ad->symb);
  psPlArray(&ad->x[ad->ix][ad->istart], &ad->x[ad->iy][ad->istart],
            ad->fx, ad->fy, ad->n);
  if (ad->errbars)
    psPlError(&ad->x[ad->ix][ad->istart], &ad->x[ad->iy][ad->istart],
              &ad->x[ad->ie][ad->istart], ad->fx, ad->fy, ad->n);
}

/****************************************************************/
/*                                                              */
/*      psFun(fd)                                               */
/*        writes function to PS file                            */
/*                                                              */
/*      fd:       Zeiger auf Struktur mit Funktionsdaten        */
/*                                                              */
/****************************************************************/

void KPl::psFun(PFUNDATA fd)
{
  if (fd->fkt) {
    psSetColData(fd->color);
    psSymbol(fd->symb);
    psPlFun(fd->xfmin, fd->xfmax, fd->dx, fd->fx, fd->fy, fd->fkt, fd->p);
  }
}

/****************************************************************/
/*                                                              */
/*      psParFun(pfd)                                           */
/*        writes parametric function to PS file                 */
/*                                                              */
/*      pfd:      Zeiger auf Struktur mit Funktionsdaten        */
/*                                                              */
/****************************************************************/

void KPl::psParFun(PPARFUNDATA pfd)
{
  if (pfd->fktx && pfd->fkty) {
    psSetColData(pfd->color);
    psSymbol(pfd->symb);
    psPlParFun(pfd->tmin, pfd->tmax, pfd->dt, pfd->fx, pfd->fy,
             pfd->fktx, pfd->px, pfd->fkty, pfd->py);
  }
}

/****************************************************************/
/*                                                              */
/*      psLegend(ld)                                            */
/*        writes legend to PS file                              */
/*                                                              */
/*      ld:       Zeiger auf Struktur mit Legendendaten         */
/*                                                              */
/****************************************************************/

void KPl::psLegend(PLEGENDDATA ld)
{
  const double f = 28.34645669;
  int n;
  double x[2], y[2], sav;

  psSetColData(ld->colsymb);
  psSymbol(ld->symb);
  if (ld->symb > 0) {
    cm2r(ld->x - 0.4 * ld->xtext, ld->y, 1.0, 1.0, &x[0], &y[0]);
    cm2r(ld->x + 0.4 * ld->xtext, ld->y, 1.0, 1.0, &x[1], &y[1]);
    n = 2;
  } else {
    cm2r(ld->x, ld->y, 1.0, 1.0, &x[0], &y[0]);
    n = 1;
  }
  psPlArray(x, y, 1.0, 1.0, n);
  psSetColFrame(ld->coltext);
  sav = psRelSize(ld->relsiz);
  psText(ld->s, f * (ld->x + ld->xtext), f * (ld->y + ld->ytext), 1);
  psRelSize(sav);
}

/****************************************************************/
/*                                                              */
/*      doPS(f)                                                 */
/*        writes PS file containing frame grid, labels, arrays, */
/*        ordinary and parametric functions                     */
/*                                                              */
/*      f:        pointer to stream opened for write access     */
/*                                                              */
/****************************************************************/

void KPl::doPS(FILE *f)
{
  int i;

  if (nItems) {
    psInit(f);
    for (i = 0; i < nItems; i++)
      switch (itd[i].ityp) {
        case FRAMEITEM:
          psFrame((PFRAMEDATA) itd[i].stru);
          break;
        case ARRAYITEM:
          psArray((PARRAYDATA) itd[i].stru);
          break;
        case FUNITEM:
          psFun((PFUNDATA) itd[i].stru);
          break;
        case PARFUNITEM:
          psParFun((PPARFUNDATA) itd[i].stru);
          break;
        case LEGENDITEM:
          psLegend((PLEGENDDATA) itd[i].stru);
      }
    psExit();
  }
}

void KPl::updStatus(int x, int y)
{
  double xcm, ycm, xr, yr;
  QString s;

  if (statusBar()->isVisible()) {
    if ((!i2cm(x, y, &xcm, &ycm)) && plot) {
      i2r(x, y, fxn, fyn, &xr, &yr);
      statusBar()->changeItem(s.sprintf("%9.4G, %9.4G", xr, yr),
                              IDD_STATUSTEXT2);
    } else
      statusBar()->changeItem("", IDD_STATUSTEXT2);
    statusBar()->changeItem(s.sprintf("%6.2f cm, %6.2f cm", xcm, ycm),
                            IDD_STATUSTEXT);
  }
}

void KPl::paintWidget()
{
  plInit(plWidget->view, plWidget->sHScrollPos, plWidget->sVScrollPos);
  factor = fZoom;
  format(bas.xf, bas.yf);
  doDraw();
  end();
}

void KPl::timerEvent(QTimerEvent *e)
{
  if (e->timerId() == idTimer)
    commandCallback(IDM_REFRESH);
}

void KPl::resizeEvent(QResizeEvent *)
{
  updateRects();
}

void KPl::slotDropEvent(KDNDDropZone *zone)
{
  QString u = zone->getURLList().first(), s;
  if (querySave())
    if (KFM::download(u, s)){
      fType = 0;
      getFile(s);
      KFM::removeTempFile(s);
    }
}

void KPl::calibrateScrollBars()
{
  bool hScroll = FALSE, vScroll = FALSE;

  QPaintDeviceMetrics pdm(plWidget);
  xPict = int(10.0 * fZoom * bas.xf * pdm.width() / pdm.widthMM() + 0.5);
  yPict = int(10.0 * fZoom * bas.yf * pdm.height() / pdm.heightMM() + 0.5);
  scrx = plWidget->width();
  scry = plWidget->height();
  if (xPict > scrx) {
    hScroll = TRUE;
    scry -= SCROLLBAR_SIZE;
  }
  if (yPict > scry) {
    vScroll = TRUE;
    scrx -= SCROLLBAR_SIZE;
    if (!hScroll)
      if (xPict > scrx) {
        hScroll = TRUE;
        scry -= SCROLLBAR_SIZE;
      }
  }
  if (hScroll) {
    plWidget->horScrollBar->setGeometry(0, scry, scrx, SCROLLBAR_SIZE);
    plWidget->horScrollBar->setRange(0, xPict - scrx);
    plWidget->horScrollBar->setSteps(10, scrx);
    plWidget->horScrollBar->setValue(plWidget->sHScrollPos);
    plWidget->horScrollBar->show();
  } else {
    plWidget->horScrollBar->hide();
    plWidget->sHScrollPos = 0;
  }
  if (vScroll) {
    plWidget->vertScrollBar->setGeometry(scrx, 0, SCROLLBAR_SIZE, scry);
    plWidget->vertScrollBar->setRange(0, yPict - scry);
    plWidget->vertScrollBar->setSteps(10, scry);
    plWidget->vertScrollBar->setValue(yPict - scry - plWidget->sVScrollPos);
    plWidget->vertScrollBar->show();
  } else {
    plWidget->vertScrollBar->hide();
    plWidget->sVScrollPos = 0;
  }
  plWidget->view->resize(scrx, scry);
}

void KPl::shrink()
{
  calibrateScrollBars();
  if ((xPict < scrx) || (yPict < scry)) {
    int no = 1 + (toolBar()->sizeHint().width() - 1) / width();
    int w = min(xPict, scrx) - scrx + width() + 2;
    int n = 1 + (toolBar()->sizeHint().width() - 1) / w;
    resize(w, min(yPict, scry) - scry + height() +
           menuBar()->heightForWidth(w) - menuBar()->height() +
           (n - no) * (toolBar()->height() / no) + 2);
  }
}

bool KPl::queryExit()
{
  int i;

  if (changed)
    do {
      i = KMsgBox::yesNoCancel(this, i18n("Warning"),
                               i18n("The current plot has been changed.\n"
                                    "Save file?"),
                               KMsgBox::EXCLAMATION);
      if (i == 3)
        return FALSE;
      if (i == 2)
        break;
    } while (!save(IDM_SAVE));
  if (optionsPopup->isItemChecked(IDM_SAVEATEND))
    saveOptions();
  return TRUE;
}

void KPl::saveOptions()
{
  KConfig *config = kapp->getConfig();
  config->setGroup(kapp->appName());
  config->writeEntry("Directory", currentDir);
  config->writeEntry("Width", width());
  config->writeEntry("Height", height());
  config->writeEntry("ToolBar", optionsPopup->isItemChecked(IDM_TOOLBAR));
  config->writeEntry("StatusBar", optionsPopup->isItemChecked(IDM_STATUSBAR));
  config->writeEntry("SaveAtEnd", optionsPopup->isItemChecked(IDM_SAVEATEND));
  config->writeEntry("AutoPlot", optionsPopup->isItemChecked(IDM_AUTOPLOT));
  config->writeEntry("AutoLetX", aut.sxAuto);
  config->writeEntry("AutoLetY", aut.syAuto);
  config->writeEntry("AutoLetH", aut.shAuto);
  config->writeEntry("AutoIndexX", aut.ixAuto);
  config->writeEntry("AutoIndexY", aut.iyAuto);
  config->writeEntry("AutoIndexE", aut.ieAuto);
  config->writeEntry("AutoNorm", aut.autoNorm);
  config->writeEntry("AutoErrBar", aut.autoErr);
  config->writeEntry("AutoGrid", aut.autoGrid);
  config->writeEntry("AutoPathHeader", aut.autohPath);
  config->writeEntry("AutoSymbol", aut.autoSymb);
  config->writeEntry("AutoX0", aut.x0Auto);
  config->writeEntry("AutoY0", aut.y0Auto);
  config->writeEntry("AutoXl", aut.xlAuto);
  config->writeEntry("AutoYl", aut.ylAuto);
  config->writeEntry("FormatX", bas.xf);
  config->writeEntry("FormatY", bas.yf);
  config->writeEntry("Zoom", fZoom);
  config->writeEntry("ColorFrame", bas.colFrame);
  config->writeEntry("ColorGrid", bas.colGrid);
  config->writeEntry("ColorData", bas.colData);
}

FILE *KPl::openFile(QString *fileBuf, QString fileName)
{
  FILE *f = 0; 
  *fileBuf = KFileDialog::getSaveFileName(currentDir, fileName);
  if (!fileBuf->isEmpty()) {
    QFileInfo fi(*fileBuf);
    if (fi.exists())
      if (KMsgBox::yesNo(this, i18n("Warning"),
                         i18n("A file with this name exists already.\n"
                              "Do you want to overwrite it?"),
                         KMsgBox::EXCLAMATION) != 1)
        return 0;
    f = fopen(*fileBuf, "w");
  }
  return f;
}

bool KPl::querySave(void)
{
  if (changed)
    do {
      int i = KMsgBox::yesNoCancel(this, i18n("Warning"),
                                   i18n("The current plot has been changed.\n"
                                        "Save file?"),
                                   KMsgBox::EXCLAMATION);
      if (i == 3)
        return FALSE;
      if (i == 2)
        break;
    } while (!save(IDM_SAVE));
  return TRUE;
}

void KPl::open(int iType)
{
  if (querySave()) {
    QString s = KFileDialog::getOpenFileName(currentDir,
      iType == DATAFILE ? "*.dat" : "*.plo");
    if (!s.isEmpty()){
      fType = iType;
      getFile(s);
    }
  }
}

bool KPl::save(int id)
{
  FILE *f = 0;
  if (!currentPath.isEmpty && (id == IDM_SAVE) && (fType == PLOTFILE))
    f = fopen(currentPath, "w");
  else {
    QString s;
    f = openFile(&s, "*.plo");
    if (f) {
      currentPath = s;
      QFileInfo fi(s);
      currentDir = fi.dirPath(TRUE);
      fType = PLOTFILE;
    }
  }
  if (f) {
    writePlo(f);
    changed = 0;
    updWin();
  }
  return (f);
}
 
void KPl::slotScrollLeft()
{
  plWidget->horScrollBar->subtractLine();
}

void KPl::slotScrollRight()
{
  plWidget->horScrollBar->addLine();
}

void KPl::slotScrollUp()
{
  plWidget->vertScrollBar->subtractLine();
}

void KPl::slotScrollDown()
{
  plWidget->vertScrollBar->addLine();
}

void KPl::slotScrollPageUp()
{
  plWidget->vertScrollBar->subtractPage();
}

void KPl::slotScrollPageDown()
{
  plWidget->vertScrollBar->addPage();
}

void KPl::slotScrollPageLeft()
{
  plWidget->horScrollBar->subtractPage();
}

void KPl::slotScrollPageRight()
{
  plWidget->horScrollBar->addPage();
}

void KPl::slotScrollHomeLeft()
{
  plWidget->horScrollBar->setValue(0);
}

void KPl::slotScrollHomeDown()
{
  plWidget->vertScrollBar->setValue(plWidget->vertScrollBar->maxValue());
}

void KPl::slotScrollEndRight()
{
  plWidget->horScrollBar->setValue(plWidget->horScrollBar->maxValue());
}

void KPl::slotScrollEndUp()
{
  plWidget->vertScrollBar->setValue(0);
}

void KPl::slotNew()
{
  if (changed)
    if (!querySave())
      return;
  freeAllItd();
  currentPath = "";
  updWin();
  update();
  slotItems();
}

void KPl::slotItems()
{
  ItemDlg dlg(this);
  if (dlg.exec()) {
    changed = TRUE;
    updWin();
  }
}

void KPl::slotOpenPlot()
{
  open(PLOTFILE);
}

void KPl::slotOpenData()
{
  open(DATAFILE);
}

void KPl::slotSave()
{
  save(IDM_SAVE);
}

void KPl::slotSaveAs()
{
  save(IDM_SAVEAS);
}

void KPl::slotPrint()
{
  QPrinter p;
  if (p.setup(this)) {
    plInit(&p); 
    doDraw();
    end();
  }
}

void KPl::slotClose()
{
  if(queryExit())
    if(windowList.count() > 1)
      close(TRUE);
    else {
      deleteAll();
      kapp->quit();
    }
}

void KPl::slotExit()
{
  bool frameChanged = FALSE;
  KPl *kpl;
  for (kpl = windowList.first(); kpl != 0; kpl = windowList.next())
    if (kpl->changed)
      frameChanged = TRUE;
  if (!frameChanged) {
    deleteAll();
    kapp->quit();
  } else
    if (KMsgBox::yesNo(this, i18n("Warning"),
                       i18n("There are windows with modified plots.\n"
                            "If you exit now, you will loose those changes.\n"
                            "Quit anyways?"), KMsgBox::EXCLAMATION) == 1) {
      deleteAll();
      kapp->quit();
    }
}

void KPl::slotHelp()
{
  kapp->invokeHTMLHelp("", 0);
}

void KPl::commandCallback(int id)
{
  switch (id) {
    case IDM_OPENDAT:
      open(DATAFILE);
      break;
    case IDM_OPENPLO:
      open(PLOTFILE);
      break;
    case IDM_NEW:
      slotNew();
      break;
    case IDM_ITEMS:
      slotItems();
      break;
    case IDM_TOOLBAR:
    case IDM_STATUSBAR:
      optionsPopup->setItemChecked(id, !optionsPopup->isItemChecked(id));
      if (id == IDM_TOOLBAR)
        enableToolBar();
      else
        enableStatusBar();
      break;
    case IDM_AUTOPLOT:
      plot = !optionsPopup->isItemChecked(id);
      optionsPopup->setItemChecked(id, plot);
      if (!plot)
        break;
    case IDM_PLOT:
      plot = nItems;
      if (plot)
        update();
      break;
    case IDM_PRINT:
      slotPrint();
      break;
    case IDM_PS:
      {
        QString s;
        FILE *f = 0;
        if ((f = openFile(&s, "*.ps"))) {
          doPS(f);
          QFileInfo fi(s);
          currentDir = fi.dirPath(TRUE);
        }
      }
      break;
    case IDM_SAVE:
    case IDM_SAVEAS:
      save(id);
      break;
    case IDM_AUTOSETTINGS:
      {
        autoStruct a = aut;
        AutoDlg dlg(this);
        dlg.setWidgets(a);
        if (dlg.exec() == dlg.Accepted) {
          aut = dlg.getValues();
        }
      }
      break;
    case IDM_SAVEOPTIONS:
      saveOptions();
      break;
    case IDM_SAVEATEND:
      optionsPopup->setItemChecked(IDM_SAVEATEND,
        !optionsPopup->isItemChecked(IDM_SAVEATEND));
      break;
    case IDM_ZOOM:
      {
        ZoomDlg dlg(this);
        dlg.factor->setValue((int) (100 * fZoom + 0.5));
        if (dlg.exec() == dlg.Accepted) {
          fZoom = 0.01 * dlg.factor->value();
          shrink();
        }
      }
      break;
    case IDM_BASICSETTINGS:
      {
        basicStruct b = bas;
        BasicDlg dlg(this);
        dlg.setWidgets(b);
        if (dlg.exec() == dlg.Accepted) {
          bas = dlg.getValues();
          shrink();
        }
      }
      break;
    case IDM_REFRESH:
      optionsPopup->setItemChecked(IDM_AUTOPLOT, TRUE);
      getFile(currentPath);
      break;
    case IDM_AUTOREFRESH:
      {
        RefreshDlg dlg(this);
        dlg.setWidgets(tRefresh, timerActive);
        if (dlg.exec() == dlg.Accepted) {
          int tRef;
          bool act;
          dlg.getValues(&tRef, &act);
          if (timerActive && ((tRef != tRefresh) || (act != timerActive)))
            killTimer(idTimer);
          if (act && ((tRef != tRefresh) || (act != timerActive))) {
            idTimer = startTimer(1000 * tRef);
            if (!idTimer) {
              KMsgBox::message(this, i18n("Error"),
                               i18n("No timer available!"), 2);
              act = 0;
            }
          }
          timerActive = act;
          tRefresh = tRef;
        }
      }
      break;
    case IDM_NEWWINDOW:
      {
        KPl *kpl = new KPl();
        kpl->show();
        break;
      }
    case IDM_CLOSE:
      slotClose();
      break;
    case IDM_EXIT:
      slotExit();
      break;
  }
}

KPl::KPl()
{
  QString s;

  windowList.setAutoDelete(FALSE);
  windowList.append(this);
  changed = FALSE;
  plot = nItems = fType = nyLeg = idTimer = timerActive = 0;
  tRefresh = 5;
  fxn = fyn = 1.0;
  itd = 0;
  currentPath = "";
  KConfig *config = kapp->getConfig();
  config->setGroup(kapp->appName());
  currentDir = config->readEntry("Directory",
                                 s.sprintf("%s/%s",
                                           (const char *) kapp->kde_datadir(),
                                           (const char *) kapp->appName()));
  strcpy(aut.sxAuto, config->readEntry("AutoLetX", "x"));
  strcpy(aut.syAuto, config->readEntry("AutoLetY", "y"));
  strcpy(aut.shAuto, config->readEntry("AutoLetH", ""));
  aut.ixAuto = config->readNumEntry("AutoIndexX", 0);
  aut.iyAuto = config->readNumEntry("AutoIndexY", 1);
  aut.ieAuto = config->readNumEntry("AutoIndexE", 2);
  aut.autoNorm = config->readBoolEntry("AutoNorm", TRUE);
  aut.autoErr = config->readBoolEntry("AutoErrBar", FALSE);
  aut.autoGrid = config->readNumEntry("AutoGrid", 2);
  aut.autohPath = config->readBoolEntry("AutoPathHeader", TRUE);
  aut.autoSymb = config->readNumEntry("AutoSymbol", 1);
  aut.x0Auto = config->readDoubleNumEntry("AutoX0", 4.0);
  aut.y0Auto = config->readDoubleNumEntry("AutoY0", 3.0);
  aut.xlAuto = config->readDoubleNumEntry("AutoXl", 15.0);
  aut.ylAuto = config->readDoubleNumEntry("AutoYl", 10.0);
  bas.xf = config->readDoubleNumEntry("FormatX", 21.0);
  bas.yf = config->readDoubleNumEntry("FormatY", 29.7);
  fZoom = config->readDoubleNumEntry("Zoom", 1.0);
  bas.colFrame = config->readLongNumEntry("ColorFrame", 0);
  bas.colGrid = config->readLongNumEntry("ColorGrid", 0);
  bas.colData = config->readLongNumEntry("ColorData", 0);
  filePopup = new QPopupMenu();
  filePopup->insertItem(i18n("&New Plot File..."), IDM_NEW);
  filePopup->insertItem(i18n("&Open Plot File..."), IDM_OPENPLO);
  filePopup->insertItem(i18n("Open &Data File..."), IDM_OPENDAT);
  filePopup->insertSeparator();
  filePopup->insertItem(i18n("&Save"), IDM_SAVE);
  filePopup->insertItem(i18n("Save &As..."), IDM_SAVEAS);
  filePopup->insertSeparator();
  filePopup->insertItem(i18n("P&lot"), IDM_PLOT);
  filePopup->insertItem(i18n("PostScript O&utput..."), IDM_PS);
  filePopup->insertSeparator();
  filePopup->insertItem(i18n("&Print..."), IDM_PRINT);
  filePopup->insertSeparator();
  filePopup->insertItem(i18n("New &Window"), IDM_NEWWINDOW);
  filePopup->insertItem(i18n("&Close Window"), IDM_CLOSE);
  filePopup->insertSeparator();
  filePopup->insertItem(i18n("E&xit"), IDM_EXIT);
#ifndef USE_KACCEL
  KStdAccel *keys = new KStdAccel(config);
  filePopup->setAccel(keys->openNew(), IDM_NEW);
  filePopup->setAccel(keys->open(), IDM_OPENPLO);
  filePopup->setAccel(keys->save(), IDM_SAVE);
  filePopup->setAccel(keys->print(), IDM_PRINT);
  filePopup->setAccel(keys->close(), IDM_CLOSE);
  filePopup->setAccel(keys->quit(), IDM_EXIT);
#else
  KAccel *keys = new KAccel(this);
  keys->connectItem(KAccel::New, this, SLOT(slotNew()));
  keys->changeMenuAccel(filePopup, IDM_NEW, KAccel::New );
  keys->connectItem(KAccel::Open, this, SLOT(slotOpenPlot()));
  keys->changeMenuAccel(filePopup, IDM_OPENPLO, KAccel::Open);
  keys->connectItem(KAccel::Save, this, SLOT(slotSave()));
  keys->changeMenuAccel(filePopup, IDM_SAVE, KAccel::Save);
  keys->connectItem(KAccel::Print, this, SLOT(slotPrint()));
  keys->changeMenuAccel(filePopup, IDM_PRINT, KAccel::Print);
  keys->connectItem(KAccel::Close, this, SLOT(slotClose()));
  keys->changeMenuAccel(filePopup, IDM_CLOSE, KAccel::Close);
  keys->connectItem(KAccel::Quit, this, SLOT(slotExit()));
  keys->changeMenuAccel(filePopup, IDM_EXIT, KAccel::Quit);
#endif
  menuBar()->insertItem(i18n("&File"), filePopup);
  menuBar()->insertItem(i18n("&Edit"), IDM_ITEMS);
  displayPopup = new QPopupMenu();
  displayPopup->insertItem(i18n("&Zoom..."), IDM_ZOOM);
  displayPopup->insertSeparator();
  displayPopup->insertItem(i18n("&Reload"), IDM_REFRESH);
  displayPopup->setAccel(Key_F5, IDM_REFRESH);
  displayPopup->insertItem(i18n("&Reload periodically..."), IDM_AUTOREFRESH);
  menuBar()->insertItem(i18n("&View"), displayPopup);
  optionsPopup = new QPopupMenu();
  optionsPopup->insertItem(i18n("Show &Tool Bar"), IDM_TOOLBAR);
  optionsPopup->insertItem(i18n("Show Stat&us Bar"), IDM_STATUSBAR);
  optionsPopup->insertItem(i18n("&Autoplot"), IDM_AUTOPLOT);
  optionsPopup->insertItem(i18n("Save Options At &End"), IDM_SAVEATEND);
  optionsPopup->insertSeparator();
  optionsPopup->insertItem(i18n("Configure &Basics..."), IDM_BASICSETTINGS);
  optionsPopup->insertItem(i18n("&Configure Autoplot..."), IDM_AUTOSETTINGS);
  optionsPopup->insertSeparator();
  optionsPopup->insertItem(i18n("&Save Options"), IDM_SAVEOPTIONS);
  optionsPopup->setCheckable(TRUE);
  menuBar()->insertItem(i18n("&Options"), optionsPopup);
  menuBar()->insertSeparator();
  menuBar()->insertItem(i18n("&Help"),
                        kapp->getHelpMenu(TRUE,
                                          "kpl " VERSION 
                                           "\n\n 1999 "
                                           "Werner Stille\n"
                                           "stille@uni-freiburg.de"));
  connect(menuBar(), SIGNAL(activated(int)), SLOT(commandCallback(int)));
  statusBar()->insertItem(" 10.00 cm,  15.00 cm ", IDD_STATUSTEXT);
  statusBar()->insertItem("", IDD_STATUSTEXT2);
  statusBar()->changeItem("", IDD_STATUSTEXT);
  if (config->readBoolEntry("StatusBar", TRUE))
    optionsPopup->setItemChecked(IDM_STATUSBAR, TRUE);
  else
    enableStatusBar(KStatusBar::Hide);
  optionsPopup->setItemChecked(IDM_SAVEATEND,
                               config->readBoolEntry("SaveAtEnd"));
  optionsPopup->setItemChecked(IDM_AUTOPLOT,
                               config->readBoolEntry("AutoPlot"));
  toolBar()->insertButton(kapp->getIconLoader()->loadIcon("filenew.xpm"),
                          IDM_NEW, TRUE, i18n("New Plot File..."));
  toolBar()->insertButton(kapp->getIconLoader()->loadIcon("fileopen.xpm"),
                          IDM_OPENPLO, TRUE, i18n("Open Plot File..."));
  toolBar()->insertButton(kapp->getIconLoader()->loadIcon("fileopen.xpm"),
                          IDM_OPENDAT, TRUE, i18n("Open Data File..."));
  toolBar()->insertSeparator();
  toolBar()->insertButton(kapp->getIconLoader()->loadIcon("filefloppy.xpm"),
                          IDM_SAVE, TRUE, i18n("Save"));
  toolBar()->insertSeparator();
  toolBar()->insertButton(kapp->getIconLoader()->loadIcon("fileprint.xpm"),
                          IDM_PRINT, TRUE, i18n("Print"));
  toolBar()->insertSeparator();
  toolBar()->insertButton(kapp->getIconLoader()->loadIcon("viewzoom.xpm"),
                          IDM_ZOOM, TRUE, i18n("Zoom..."));
  toolBar()->insertSeparator();
  toolBar()->insertButton(kapp->getIconLoader()->loadIcon("reload.xpm"),
                          IDM_REFRESH, TRUE, i18n("Reload"));
  toolBar()->insertSeparator();
  toolBar()->insertButton(kapp->getIconLoader()->loadIcon("help.xpm"),
                          0, SIGNAL(clicked()), this,
                          SLOT(slotHelp()), TRUE, i18n("Help"));
  connect(toolBar(), SIGNAL(clicked(int)), SLOT(commandCallback(int)));
  if (config->readBoolEntry("ToolBar", TRUE))
    optionsPopup->setItemChecked(IDM_TOOLBAR, TRUE);
  else
    enableToolBar(KToolBar::Hide);
  setMinimumSize(120, 80);
  plWidget = new KPlWidget(this);
  setView(plWidget);
  resize(config->readNumEntry("Width", 600),
         config->readNumEntry("Height", 400));
  connect(new KDNDDropZone(this, DndURL), 
	  SIGNAL(dropAction(KDNDDropZone *)), 
	  SLOT(slotDropEvent(KDNDDropZone *)));
#ifndef USE_KACCEL
  keys->registerWidget("kpl", this);
  keys->addKey("ScrollLeft", Key_Left);
  keys->addKey("ScrollRight", Key_Right);
  keys->addKey("ScrollUp", Key_Up);
  keys->addKey("ScrollDown", Key_Down);
  keys->addKey("ScrollPageUp", Key_PageUp);
  keys->addKey("ScrollPageDown", Key_PageDown);
  keys->addKey("ScrollPageLeft", CTRL + Key_PageUp);
  keys->addKey("ScrollPageRight", CTRL + Key_PageDown);
  keys->addKey("ScrollHomeLeft", CTRL + Key_Home);
  keys->addKey("ScrollHomeDown", Key_Home);
  keys->addKey("ScrollEndRight", CTRL + Key_End);
  keys->addKey("ScrollEndUp", Key_End);
  keys->connectFunction("kpl", "ScrollLeft", this, SLOT(slotScrollLeft()));
  keys->connectFunction("kpl", "ScrollRight", this, SLOT(slotScrollRight()));
  keys->connectFunction("kpl", "ScrollUp", this, SLOT(slotScrollUp()));
  keys->connectFunction("kpl", "ScrollDown", this, SLOT(slotScrollDown()));
  keys->connectFunction("kpl", "ScrollPageUp", this, SLOT(slotScrollPageUp()));
  keys->connectFunction("kpl", "ScrollPageDown", this,
                        SLOT(slotScrollPageDown()));
  keys->connectFunction("kpl", "ScrollPageLeft", this,
                        SLOT(slotScrollPageLeft()));
  keys->connectFunction("kpl", "ScrollPageRight", this,
                        SLOT(slotScrollPageRight()));
  keys->connectFunction("kpl", "ScrollHomeLeft", this,
                        SLOT(slotScrollHomeLeft()));
  keys->connectFunction("kpl", "ScrollHomeDown", this,
                        SLOT(slotScrollHomeDown()));
  keys->connectFunction("kpl", "ScrollEndRight", this,
                        SLOT(slotScrollEndRight()));
  keys->connectFunction("kpl", "ScrollEndUp", this, SLOT(slotScrollEndUp()));
#else
  keys->insertItem(i18n("Scroll Left"), "ScrollLeft", Key_Left);
  keys->insertItem(i18n("Scroll Right"), "ScrollRight", Key_Right);
  keys->insertItem(i18n("Scroll Up"), "ScrollUp", Key_Up);
  keys->insertItem(i18n("Scroll Down"), "ScrollDown", Key_Down);
  keys->insertItem(i18n("Scroll Page Up"), "ScrollPageUp", Key_PageUp);
  keys->insertItem(i18n("Scroll Page Down"), "ScrollPageDown", Key_PageDown);
  keys->insertItem(i18n("Scroll Page Left"), "ScrollPageLeft",
                   CTRL + Key_PageUp);
  keys->insertItem(i18n("Scroll Page Right"), "ScrollPageRight",
                   CTRL + Key_PageDown);
  keys->insertItem(i18n("Scroll to Left Margin"), "ScrollHomeLeft",
                   CTRL + Key_Home);
  keys->insertItem(i18n("Scroll to Bottom Margin"), "ScrollHomeDown",
                   Key_Home);
  keys->insertItem(i18n("Scroll to Right Margin"), "ScrollEndRight",
                   CTRL + Key_End);
  keys->insertItem(i18n("Scroll to Top Margin"), "ScrollEndUp", Key_End);
  keys->connectItem("ScrollLeft", this, SLOT(slotScrollLeft()));
  keys->connectItem("ScrollRight", this, SLOT(slotScrollRight()));
  keys->connectItem("ScrollUp", this, SLOT(slotScrollUp()));
  keys->connectItem("ScrollDown", this, SLOT(slotScrollDown()));
  keys->connectItem("ScrollPageUp", this, SLOT(slotScrollPageUp()));
  keys->connectItem("ScrollPageDown", this, SLOT(slotScrollPageDown()));
  keys->connectItem("ScrollPageLeft", this, SLOT(slotScrollPageLeft()));
  keys->connectItem("ScrollPageRight", this, SLOT(slotScrollPageRight()));
  keys->connectItem("ScrollHomeLeft", this, SLOT(slotScrollHomeLeft()));
  keys->connectItem("ScrollHomeDown", this, SLOT(slotScrollHomeDown()));
  keys->connectItem("ScrollEndRight", this, SLOT(slotScrollEndRight()));
  keys->connectItem("ScrollEndUp", this, SLOT(slotScrollEndUp()));
#endif
  updWin();
}

KPl::~KPl()
{
  delete filePopup;
  delete optionsPopup;
  delete displayPopup;
  delete plWidget;
  windowList.removeRef(this);
}

int main(int argc, char **argv)
{
  KApplication a(argc, argv);

  if (a.isRestored()) {
    int n = 1;
    while (KTMainWindow::canBeRestored(n)) {
      KPl *kpl = new KPl();
      kpl->KTMainWindow::restore(n++);
    }
  } else {
    KPl *kpl = new KPl();
    if (argc > 1)
      kpl->getFile(argv[1]);
    kpl->show();
  }
  return a.exec();
}
